#include <any>
#include <array>
#include <iostream>

struct reporter {
	reporter() { std::cout << "constructor\n"; }
	~reporter() { std::cout << "destructor\n"; }
	reporter(reporter&&) noexcept { std::cout << "move ctor\n"; }
	reporter(const reporter&) { std::cout << "copy ctor\n"; }
	reporter& operator=(reporter&&) noexcept {
		std::cout << "move assignment\n";
		return *this;
	}
	reporter& operator=(const reporter&) {
		std::cout << "copy assignment\n";
		return *this;
	}
};

struct thing {
	reporter r;
	// try uncommenting the following line:
	// std::array<int, 100> arr;
};


int main() {
	std::any a = 42;
	try {
		std::cout << std::any_cast<int>(a) << '\n';
	} catch (std::bad_any_cast& ex) {
		std::cout << "Exception: " << ex.what() << '\n';
	}

	// pointer version
	if (auto* ptr = std::any_cast<int>(&a)) {
		std::cout << "got int: " << *ptr << '\n';
	} else {
		std::cout << "not int\n";
	}

	std::cout << "--- created, then moved:\n";
	std::any a1 = thing{};
	std::cout << "---\n";
	a1.reset();

	std::cout << "--- created directly:\n";
	std::any a2 = std::make_any<thing>();

	std::cout << "--- copy the held object:\n";
	auto t1 = std::any_cast<thing>(a2);

	std::cout << "--- reference to the held object:\n";
	auto& ref = std::any_cast<thing&>(a2);

	std::cout << "--- move the held object:\n";
	auto t2 = std::any_cast<thing>(std::move(a2));
	std::cout << "---\n";

	auto a3 = std::make_any<thing>();

	std::cout << "--- copy any\n";
	auto clone = a3;

	std::cout << "--- move any\n";
	auto moved = std::move(a3);

	std::cout << "---\n";
}
