#include <cassert>
#include <functional>
#include <iostream>
#include <vector>

struct int_notifier {
	std::vector<std::function<void(int)>> observers;

	void add_observer(std::function<void(int)> obs) {
		assert(obs);  // obs should not be empty
		observers.push_back(std::move(obs));
	}

	void notify(int num) const {
		for (auto& obs : observers) {
			obs(num);
		}
	}
};

void print(int num) { std::cout << num << '\n'; }
void print_vec(const auto& vec) {
	char sep = '{';
	for (auto& x : vec) {
		std::cout << sep << ' ' << x;
		sep = ',';
	}
	std::cout << (vec.empty() ? "{}\n" : " }\n");
}

struct remember {
	std::vector<int> memory;
	void operator()(int num) {
		memory.push_back(num);
		print_vec(memory);
	}
};

int main() {
	int_notifier n;

	n.add_observer(print);

	int last;
	n.add_observer([&](int num) { last = num; });

	n.add_observer([sum = 0](int num) mutable {
		sum += num;
		std::cout << "Sum so far: " << sum << '\n';
	});

	remember r;
	n.add_observer(r);

	for (int i = 1; i < 10; ++i) {
		n.notify(i);
		std::cout << "last: " << last << '\n';
	}

	std::cout << "r.memory.size(): " << r.memory.size() << '\n';
	// note: the observer inside n is not r, but r's copy
	// if you want to know how to get the remember object out of
	// n.observers, look at the target() method
}
