#include <iostream>
#include <variant>
#include <vector>

struct spider {
	void describe() const {
		std::cout << "a scary spider";
	}
};
struct troll {
	void describe() const {
		std::cout << "an ugly troll";
	}
};
using monster = std::variant<spider, troll>;

struct axe {
	void describe() const {
		std::cout << "an axe";
	}
};
struct nuke {
	void describe() const {
		std::cout << "a tactical nuclear missile";
	}
};
using weapon = std::variant<axe, nuke>;

struct attack {
	void operator()(auto w, auto m) const {
		std::cout << "Using ";
		w.describe();
		std::cout << " against ";
		m.describe();
		std::cout << ":\n\t";
		apply(w, m);
	}

private:
	void apply(auto, auto) const {
		std::cout << "(generic weapon-monster interaction)\n";
	}
	void apply(axe, spider) const {
		std::cout << "You miss.\n";
	}
	void apply(axe, troll) const {
		std::cout << "You hit the troll.\n";
	}
	void apply(nuke, spider) const {
		std::cout << "The spider is annihilated.\n";
	}
	void apply(nuke, troll) const {
		std::cout << "Trolls are immune to nuclear weapons, silly.\n";
	}
};

int main() {
	std::vector<monster> monsters = { spider{}, troll{} };
	std::vector<weapon> weapons = { axe{}, nuke{} };

	for (auto w : weapons) {
		for (auto m : monsters) {
			std::visit(attack{}, w, m);
		}
	}
}
