#include <iostream>
#include <optional>

struct reporter {
	reporter() { std::cout << "constructor\n"; }
	~reporter() { std::cout << "destructor\n"; }
	reporter(reporter&&) noexcept { std::cout << "move ctor\n"; }
	reporter(const reporter&) { std::cout << "copy ctor\n"; }
	reporter& operator=(reporter&&) noexcept {
		std::cout << "move assignment\n";
		return *this;
	}
	reporter& operator=(const reporter&) {
		std::cout << "copy assignment\n";
		return *this;
	}
};

void line(bool large = false) {
	if (large) std::cout << "-----";
	std::cout << "-----------------\n";
}

int main() {
	{
		auto optr = std::make_optional<reporter>();
		line();
		auto copy = optr;
		line();
	}
	line(true);
	{
		auto optr = std::make_optional<reporter>();
		line();
		auto stolen = std::move(optr);
		line();
		std::cout << optr.has_value() << '\n';
	}
	line(true);
	{
		auto optr = std::make_optional<reporter>();
		line();
		optr = {};  // or: optr = std::nullopt; or: optr.reset();
		line();
	}
	line(true);
	{
		auto optr = std::make_optional<reporter>();
		line();
		std::optional<reporter> optr2 = reporter{};
		line();
		optr = optr2;
		line();
	}
	line(true);
	{
		std::optional<reporter> optr;
		optr.emplace();
		line();
		std::optional<reporter> optr2 = reporter{};
		line();
		optr = std::move(optr2);
		line();
		std::cout << optr2.has_value() << '\n';
	}
}
