#include <iostream>
#include <variant>

struct reporter {
	reporter() { std::cout << "constructor\n"; }
	~reporter() { std::cout << "destructor\n"; }
	reporter(reporter&&) noexcept { std::cout << "move ctor\n"; }
	reporter(const reporter&) { std::cout << "copy ctor\n"; }
	reporter& operator=(reporter&&) noexcept {
		std::cout << "move assignment\n";
		return *this;
	}
	reporter& operator=(const reporter&) {
		std::cout << "copy assignment\n";
		return *this;
	}
};

void line(bool large = false) {
	if (large) std::cout << "-----";
	std::cout << "-----------------\n";
}

int main() {
	using Var = std::variant<reporter, int>;
	{
		Var v;
		line();
		auto copy = v;
		line();
	}
	line(true);
	{
		Var v;
		line();
		auto stolen = std::move(v);
		line();
		std::cout << std::holds_alternative<reporter>(v) << '\n';
	}
	line(true);
	{
		Var v;
		line();
		v.emplace<int>(42);
		std::cout << std::get<int>(v) << '\n';
		v.emplace<0>();  // reporter
	}
	line(true);
	{
		Var v1;
		Var v2;
		v1 = v2;
	}
	line(true);
	{
		Var v1;
		Var v2;
		v1 = std::move(v2);
	}
	line(true);
	{
		Var v1;
		Var v2 = 42;
		v1 = v2;
	}
	line(true);
	{
		Var v1 = 42;
		Var v2;
		v1 = v2;
	}
}
